﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;

namespace Valentia.Kinection.DataTypes
{
    #region SkeletonJointVector CLASS
    /// <summary>
    /// Skeleton joint vector class.
    /// </summary>
    public class SkeletonJointVector
    {
        #region PROPERTIES
        /// <summary>
        /// Gets W value.
        /// </summary>
        public float W { get; private set; }

        /// <summary>
        /// Gets X value.
        /// </summary>
        public float X { get; private set; }

        /// <summary>
        /// Gets Y value.
        /// </summary>
        public float Y { get; private set; }

        /// <summary>
        /// Gets Z value.
        /// </summary>
        public float Z { get; private set; }
        #endregion

        #region CONSTRUCTOR
        /// <summary>
        /// Constrcutor.
        /// </summary>
        /// <param name="w">W</param>
        /// <param name="x"><X/param>
        /// <param name="y">Y</param>
        /// <param name="z">Z</param>
        public SkeletonJointVector(float w, float x, float y, float z)
        {
            this.W = w;
            this.X = x;
            this.Y = y;
            this.Z = z;
        }
        #endregion
    }
    #endregion

    #region SkeletonJointDepth CLASS
    /// <summary>
    /// Skeleton joint depth.
    /// </summary>
    public class SkeletonJointDepth
    {
        #region PROPERTIES
        /// <summary>
        /// Gets depth of X.
        /// </summary>
        public float DepthX { get; private set; }

        /// <summary>
        /// Gets depth of Y.
        /// </summary>
        public float DepthY { get; private set; }

        /// <summary>
        /// Gets depth value.
        /// </summary>
        public short DepthValue { get; private set; }
        #endregion

        #region CONSTRUCTOR
        /// <summary>
        /// Conctructor.
        /// </summary>
        /// <param name="depthX">Depth of X.</param>
        /// <param name="depthY">Depth of Y.</param>
        /// <param name="depthValue">Depth value.</param>
        public SkeletonJointDepth(float depthX, float depthY, short depthValue)
        {
            this.DepthX = depthX;
            this.DepthY = depthY;
            this.DepthValue = depthValue;
        }
        #endregion
    }
    #endregion

    #region SkeletonJointInfo CLASS
    /// <summary>
    /// Skeleton joint information class
    /// </summary>
    public class SkeletonJointInfo
    {
        #region PROPERTIES
        /// <summary>
        /// Gets the joint ID.
        /// </summary>
        public SkeletonJointID JointID { get; private set; }

        /// <summary>
        /// Gets the x point of the joint mapped to color video image.
        /// </summary>
        public double ColorX { get; private set; }

        /// <summary>
        /// Gets the y point of joint mapped to color video image.
        /// </summary>
        public double ColorY { get; private set; }

        /// <summary>
        /// Gets the joint vector.
        /// </summary>
        public SkeletonJointVector Vector { get; private set; }

        /// <summary>
        /// Gets the joint depth.
        /// </summary>
        public SkeletonJointDepth JointDepth { get; private set; }
        #endregion

        #region CONSTRUCTOR
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="jointID">Joint ID.</param>
        /// <param name="colorX">X point of the joint mapped to color video image.</param>
        /// <param name="colorY">Y point of joint mapped to color video image.</param>
        /// <param name="vector">Joint vector.</param>
        /// <param name="jointDepth">Joint depth.</param>
        public SkeletonJointInfo(SkeletonJointID jointID, double colorX, double colorY, SkeletonJointVector vector, SkeletonJointDepth jointDepth)
        {
            this.JointID = jointID;
            this.ColorX = colorX;
            this.ColorY = colorY;
            this.Vector = vector;
            this.JointDepth = jointDepth;
        }
        #endregion

        #region METHODS
        /// <summary>
        /// Converts this instance into System.Drawing.Point.
        /// </summary>
        /// <returns>System.Drawing.Point instance</returns>
        public System.Drawing.Point ToPoint() { return new System.Drawing.Point((int)(this.ColorX), (int)(this.ColorY)); }
        #endregion
    }
    #endregion

    #region SkeletonJointCollection CLASS
    /// <summary>
    /// Skeleton joints collection class.
    /// </summary>
    public class SkeletonJointCollection : ReadOnlyCollection<SkeletonJointInfo> //ICollection<JointPoint>
    {
        #region DATA MEMBERS
        /// <summary>
        /// Contains the list of joints with information
        /// </summary>
        private Dictionary<SkeletonJointID, SkeletonJointInfo> mappedJoints;
        #endregion

        #region PROPERTIES
        /// <summary>
        /// Gets the tracking ID of skeleton. All joints in this collection are for the same skeleton.
        /// </summary>
        public int TrackingID { get; private set; }

        /// <summary>
        /// Gets the user index of skeleton. All joints in this collection are for the same skeleton.
        /// </summary>
        public int UserIndex { get; private set; }

        /// <summary>
        /// Gets the quality of skeleton.
        /// </summary>
        public KinectSkeletonQuality Quality { get; private set; }

        /// <summary>
        /// Gets the skeleton joint by ID.
        /// </summary>
        /// <param name="ID">Skeleton joint ID</param>
        /// <returns>Skeleton joint.</returns>
        public SkeletonJointInfo this[SkeletonJointID ID] { get { return this.mappedJoints[ID]; } }
        #endregion

        #region CONSTRUCTOR
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="trackingID">Tracking ID of skeleton. All joints in this collection should be for this skeleton ID.</param>
        /// <param name="userIndex">User index of skeleton. All joints in this collection should be for this user index.</param>
        /// <param name="quality">Quality of skeleton.</param>
        /// <param name="mappedJoints">Contains the collection of joints mapped with joint ID.</param>
        public SkeletonJointCollection(int trackingID, int userIndex, KinectSkeletonQuality quality, Dictionary<SkeletonJointID, SkeletonJointInfo> mappedJoints)
            : base(new List<SkeletonJointInfo>(mappedJoints.Count))
        {
            this.mappedJoints = mappedJoints;
            this.TrackingID = trackingID;
            this.UserIndex = userIndex;
            this.Quality = quality;

            foreach (SkeletonJointInfo joint in this.mappedJoints.Values) { base.Items.Add(joint); }
        }
        #endregion
    }
    #endregion
}